"""
Created for storing version number.
The version number will be placed here based on the tag in Github using the tags_marker.py script.

Having "2.0.4" in final releases is NOT allowed.
"""
__version__="2.0.4"
